/* wydawca - automatic release submission daemon
   Copyright (C) 2007, 2008, 2009 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "sql.h"
#include "argmatch.h"
#include <mail.h>
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free
#include "obstack.h"

/* Configuration file handling (application-specific) */

static struct obstack cfg_stk;
static struct access_method default_verify_method;
static struct access_method default_gpg_key_method;
static struct access_method default_project_owner_method;
static struct access_method default_user_data_method;
static struct archive_descr default_archive_descr;

#define ISHSPACE(c) ((c)==' ' || (c)=='\t')
#define skip_ws(s)  while (*(s) && isspace (*(s))) (s)++;
#define skip_hws(s)  while (*(s) && ISHSPACE (*(s))) (s)++;
#define skip_word(s) while (*(s) && !isspace (*(s))) (s)++;

/* Return a pointer to the next word from *PSTR. Advanse *PSTR to the beginning
   of next word, or to the terminating \0, whichever comes first. */
static char *
get_word (char **pstr)
{
  char *word;
  char *str = *pstr;

  if (!str)
    {
      *pstr = "";
      return *pstr;
    }
  
  skip_ws (str);

  word = str;
  if (*word)
    {
      skip_word (str);
      
      if (*str)
	{
	  *str++ = 0;
	  skip_ws (str);
	}
    }
  *pstr = str;
  return word;
}

/* Convert a boolean from STR into its C representation. If STR is a valid
   boolean, store the result in *PVAL and return 0. Otherwise, return 1 */ 
static int
get_bool (char *str, int *pval)
{
  if (strcmp (str, "yes") == 0
      || strcmp (str, "t") == 0)
    *pval = 1;
  else if (strcmp (str, "no") == 0
	   || strcmp (str, "nil") == 0)
    *pval = 0;
  else
    {
      char *p;
      int n = strtoul (str, &p, 10);
      if (*p == 0)
	*pval = n;
      else
	return 1;
    }
  return 0;
}

/* safe_file_name: convert a file name possibly containig relative
   specs (../) into a safer form using only direct descendence.
   Strip trailing delimiter if present, unless it is the only character
   left. E.g.:

         /home/user/../smith   -->   /home/smith
	 /home/user/../..      -->   /
	 ../file               -->   NULL
*/	 
char *
safe_file_name (char *file_name)
{
  int len;
  char *p;
  
  if (!file_name)
    return file_name;

  len = strlen (file_name);

  /* Empty string is returned as is */
  if (len == 0)
    return file_name;

  /* delete trailing delimiter if any */
  if (len && file_name[len-1] == '/')
    file_name[len-1] = 0;

  /* Eliminate any ./ and /../ */
  for (p = strchr (file_name, '.'); p; p = strchr (p, '.'))
    {
      if (p[1] == '/' && (p == file_name || p[-1] == '/'))
	{
	  char *q, *s;

	  s = p + 2;
	  q = p;
	  while ((*q++ = *s++))
	    ;
	  continue;
	}
      else if (p[1] == '.' && (p[2] == 0 || p[2] == '/'))
	{
	  if (p == file_name)
	    return NULL;
	  if (p[-1] == '/')
	    /* found */
	    {
	      char *q, *s;

	      s = p + 2;

	      /* Find previous delimiter */
	      for (q = p-2; *q != '/' && q >= file_name; q--)
		;

	      if (q < file_name)
		{
		  q = file_name;
		  s++;
		}

	      /* Copy stuff */
	      p = q;
	      while ((*q++ = *s++))
		;
	      continue;
	    }
	}
      
      p++;
    }

  if (file_name[0] == 0)
    {
      file_name[0] = '/';
      file_name[1] = 0;
    }

  return file_name;
}

/* Same as safe_file_name, but returns an allocated copy. */
char *
safe_file_name_alloc (const char *file_name)
{
  char *s = xstrdup (file_name);
  char *ns = safe_file_name (s);
  if (!ns)
    free (s);
  return ns;
}


static void
cfg_syslog_tag (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  syslog_tag = strdup (val);
}

static void
cfg_syslog_print_priority (gsc_config_file_t *file, char *kw, char *val,
			   void *unused)
{
  if (get_bool (val, &syslog_include_prio))
    {
      file->error_msg (file->file_name, file->line,
		       "boolean value expected, but found `%s'", val);
      file->error_count++;
    }
}

static void
cfg_syslog_facility (gsc_config_file_t *file, char *kw, char *val,
		     void *unused)
{
  if (gsc_str_to_syslog_facility (val, &log_facility))
    {
      file->error_msg (file->file_name, file->line, "Unknown facility `%s'",
		       val);
      file->error_count++;
    }
}

static void
cfg_source (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  dp->source_dir = xstrdup (val);
}

static void
cfg_destination (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  dp->dest_dir = xstrdup (val);
}

static void
cfg_file_sweep_time (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  time_t interval;
  const char *endp;

  if (parse_time_interval (val, &interval, &endp))
    {
      file->error_msg (file->file_name, file->line,
		       "unrecognized time format (near `%s')",
		       endp);
      file->error_count++;
    }
  
  if (data)
    dp->file_sweep_time = interval;
  else
    file_sweep_time = interval;
}

static enum access_method_type
str_to_access_method (const char *str)
{
  if (strcmp (str, "sql") == 0)
    return method_sql;
  else if (strcmp (str, "builtin") == 0)
    return method_builtin;
  else if (strcmp (str, "external") == 0)
    return method_external;
  else
    return method_none;
}

static void
_cfg_access_method (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct access_method *ap = data;
  char *word = get_word (&val);

  if (*word == 0)
    {
      file->error_msg (file->file_name, file->line,
		       "access method not specified");
      file->error_count++;
      return;
    }
  ap->type = str_to_access_method (word);
  switch (ap->type)
    {
    case method_none:
      file->error_msg (file->file_name, file->line,
		       "unknown access method: %s", word);
      file->error_count++;
      return;
      
    case method_sql:
      word = get_word (&val);
      if (!sql_connection_exists_p (word))
	{
	  file->error_msg (file->file_name, file->line,
			   "SQL connection `%s' not declared", word);
	  ap->type = method_none;
	  return;
	}
      ap->param[0] = xstrdup (word);
      ap->param[1] = xstrdup (val);
      break;
      
    case method_builtin:
    case method_external:
      ap->param[0] = xstrdup (val);
      break;
    }
}

static void
cfg_verify_user (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  struct access_method *method;
  if (dp)
    method = dp->verify_method = method_new (method_builtin);
  else
    method = &default_verify_method;
  
  _cfg_access_method (file, kw, val, method);
}

static void
cfg_gpg_key (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  struct access_method *method;
  if (dp)
    method = dp->gpg_key_method = method_new (method_builtin);
  else
    method = &default_gpg_key_method;
  _cfg_access_method (file, kw, val, method);
}

static void
cfg_project_owner (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  struct access_method *method;
  if (dp)
    method = dp->project_owner_method = method_new (method_builtin);
  else
    method = &default_project_owner_method;
  _cfg_access_method (file, kw, val, method);
}

static void
cfg_user_data (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  struct access_method *method;
  if (dp)
    method = dp->user_data_method = method_new (method_builtin);
  else
    method = &default_user_data_method;
  _cfg_access_method (file, kw, val, method);
}

static char const * const backup_args[] =
{
  /* In a series of synonyms, present the most meaningful first, so
     that argmatch_valid be more readable. */
  "none", "off",
  "simple", "never",
  "existing", "nil",
  "numbered", "t",
  NULL
};

static const enum backup_type backup_types[] =
{
  no_backups, no_backups,
  simple_backups, simple_backups,
  numbered_existing_backups, numbered_existing_backups,
  numbered_backups, numbered_backups
};

/* Ensure that these two vectors have the same number of elements,
   not counting the final NULL in the first one.  */
ARGMATCH_VERIFY (backup_args, backup_types);

static enum backup_type
get_backup_version (gsc_config_file_t *file, const char *ctx,
		    const char *version)
{
  if (version == 0 || *version == 0)
    return numbered_existing_backups;
  else
    {
      ptrdiff_t d = ARGMATCH (version, backup_args, backup_types);
      if (d == -1)
	{
	  if (ctx)
	    file->error_msg (file->file_name, file->line,
			     "%s: invalid backup type `%s'", ctx, version);
	  else
	    file->error_msg (file->file_name, file->line,
			     "invalid backup type `%s'", version);
	  file->error_count++;
	  return no_backups;
	}
      return backup_types[d];
    }
}

static void
cfg_archive (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct directory_pair *dp = data;
  struct archive_descr *arch = dp ? &dp->archive : &default_archive_descr;
  char *type = get_word (&val);
  char *word;

  if (*type == 0)
    {
      file->error_msg (file->file_name, file->line, "missing archive type");
      file->error_count++;
      return;
    }
  
  if (strcmp (type, "none") == 0)
    {
      arch->type = archive_none;
      if (*val)
	file->error_msg (file->file_name, file->line,
		      "warning: rest of line ignored for archive type `none'");
    }
  else
    {
      word = get_word (&val);
      if (*word == 0)
	{
	  file->error_msg (file->file_name, file->line, "missing file name");
	  file->error_count++;
	  arch->type = archive_none;
	  return;
	}
      arch->name = safe_file_name (xstrdup (word));
      if (!arch->name)
        {
          file->error_msg (file->file_name, file->line,
                           "invalid archive name: %s", word);
	  file->error_count++;
          return;
        }                    
      
      word = get_word (&val);
      
      if (strcmp (type, "tar") == 0)
	{
	  arch->type = archive_tar;
	  if (*word)
	    file->error_msg (file->file_name, file->line,
			     "warning: junk after the archive name ignored");
	}
      else if (strcmp (type, "directory") == 0)
	{
	  arch->type = archive_directory;
	  if (*word)
	    {
	      arch->backup_type = get_backup_version (file, NULL, word);
	      if (*val)
		file->error_msg (file->file_name, file->line,
			     "warning: junk after the backup type ignored");
	    }
	  else
	    {
	      arch->backup_type =
		get_backup_version (file,
				    "VERSION_CONTROL environment variable",
				    getenv ("VERSION_CONTROL"));
	    }
	}
      else
	{
	  file->error_msg (file->file_name, file->line,
			   "unknown archive type: %s", word);
	  file->error_count++;
	}
    }
}
  
static void
cfg_directory (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  int ec;
  char *word;
  struct directory_pair dpair;
  static struct gsc_config_keyword keywords[] = {
    { "include", gsc_config_include_handler },
    { "source", cfg_source },
    { "destination", cfg_destination },
    { "file-sweep-time", cfg_file_sweep_time },
    { "verify-user", cfg_verify_user },
    { "gpg-key", cfg_gpg_key },
    { "project-owner", cfg_project_owner },
    { "user-data", cfg_user_data },
    { "archive", cfg_archive },
    { NULL }
  };

  memset (&dpair, 0, sizeof dpair);
  word = get_word (&val);
  if (!word)
    dpair.url = NULL;
  else
    dpair.url = xstrdup (word);
  if (val && val[0])
    file->error_msg (file->file_name, file->line,
		     "warning: junk in line ignored");
  
  dpair.file_sweep_time = file_sweep_time;
  dpair.verify_method = &default_verify_method;
  dpair.gpg_key_method = &default_gpg_key_method;
  dpair.project_owner_method = &default_project_owner_method;
  dpair.user_data_method = &default_user_data_method;
  dpair.archive = default_archive_descr;

  gsc_config_parse_block (file, &dpair, keywords, "end");

  if (!dpair.source_dir)
    {
      file->error_msg (file->file_name, file->line, "source is not given");
      file->error_count++;
    }
  else if (!dpair.dest_dir)
    {
      file->error_msg (file->file_name, file->line,
		       "destination is not given");
      file->error_count++;
    }
  else if (test_dir (dpair.source_dir, &ec))
    {
      if (ec)
	file->error_msg (file->file_name, file->line,
			 "cannot access %s: %s",
			 dpair.source_dir, strerror (ec));
      else
	file->error_msg (file->file_name, file->line,
			 "%s is not a directory",
			 dpair.source_dir);
      file->error_count++;
    }
  else if (test_dir (dpair.dest_dir, &ec))
    {
      if (ec)
	file->error_msg (file->file_name, file->line,
			 "cannot access %s: %s",
			 dpair.dest_dir, strerror (ec));
      else
	file->error_msg (file->file_name, file->line,
			 "%s is not a directory",
			 dpair.dest_dir);
      file->error_count++;
    }
  else if (dpair.verify_method->type != method_none
	   && dpair.gpg_key_method->type != method_none)
    {
      /* FIXME: only sql supported so far */
      if (dpair.verify_method->type == method_builtin
	  && dpair.gpg_key_method->type != method_builtin)
	{
	  file->error_msg (file->file_name, file->line,
			   "Sorry, method `builtin' is not yet supported");
	  file->error_count++;
	}
      else if (dpair.verify_method->type == method_external
	       && dpair.gpg_key_method->type != method_external)
	{
	  file->error_msg (file->file_name, file->line,
			   "Sorry, method `external' is not yet supported");
	  file->error_count++;
	}
	
      register_directory_pair (&dpair);
    }
}

static void
cfg_sql_host (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct sqlconn *pconn = data;
  char *p = strchr (val, ':');
  if (p)
    {
      *p++ = 0;
      if (p[0] == '/')
	{
	  pconn->socket = xstrdup (p);
	  pconn->host = xstrdup ("localhost");
	}
      else
	{
	  char *end;
	  unsigned long n = strtoul (p, &end, 10);
	  if (*end)
	    {
	      file->error_msg (file->file_name, file->line,
			       "invalid port number (near %s)", end);
	      file->error_count++;
	      return;
	    }
	  if (n == 0 || n > USHRT_MAX)
	    {
	      file->error_msg (file->file_name, file->line,
			       "port number out of range 1..%d", USHRT_MAX);
	      file->error_count++;
	      return;
	    }
	  pconn->port = n;
	  /* Save host name */
	  pconn->host = xstrdup (val);
	}
    }
  else
    pconn->host = xstrdup (val);
}

static void
cfg_sql_database (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct sqlconn *pconn = data;
  pconn->database = xstrdup (val);
}

static void
cfg_sql_user (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct sqlconn *pconn = data;
  pconn->user = xstrdup (val);
}

static void
cfg_sql_password (gsc_config_file_t *file, char *kw, char *val, void *data)
{
  struct sqlconn *pconn = data;
  pconn->password = xstrdup (val);
}

static void
cfg_sql (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  struct sqlconn sql;
  static struct gsc_config_keyword keywords[] = {
    { "include", gsc_config_include_handler },
    { "host", cfg_sql_host },
    { "database", cfg_sql_database },
    { "user", cfg_sql_user },
    { "password", cfg_sql_password },
    { NULL }
  };

  memset (&sql, 0, sizeof sql);
  if (!val)
    {
      file->error_msg (file->file_name, file->line,
		       "missing block identifier");
      file->error_count++;
    }
  else
    sql.ident = xstrdup (val);
  
  gsc_config_parse_block (file, &sql, keywords, "end");
  if (sql.ident)
    sql_register_conn (&sql);
}

static void
cfg_umask (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  char *p;
  mode_t m = strtoul (val, &p, 8) & 0777;
  if (*p)
    {
      file->error_msg (file->file_name, file->line, "invalid umask (near %s)",
		       p);
      file->error_count++;
      return;
    }
  umask (m);
}

static void
cfg_tar_program (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  char *word = safe_file_name (get_word (&val));

  if (!word || word[0] != '/')
    {
      file->error_msg (file->file_name, file->line,
		       "must be an absolute file name");
      file->error_count++;
      return;
    }

  tar_command_name = xstrdup (word);
  if (*val)
    file->error_msg (file->file_name, file->line,
		     "warning: junk after the program name ignored");
}

static const char *stat_args[] = {
  "errors",
  "warnings",
  "bad-signatures",
  "access-violations",
  "complete-triplets",
  "incomplete-triplets",
  "bad-triplets",
  "expired-triplets",
  "triplet-success",
  "uploads",
  "archives",
  "symlinks",
  "rmsymlinks",
  NULL
};

static unsigned long stat_types[] = {
  STAT_ERRORS,
  STAT_WARNINGS,
  STAT_BAD_SIGNATURE,
  STAT_ACCESS_VIOLATIONS,
  STAT_COMPLETE_TRIPLETS,
  STAT_INCOMPLETE_TRIPLETS,
  STAT_BAD_TRIPLETS,
  STAT_EXPIRED_TRIPLETS,
  STAT_TRIPLET_SUCCESS,
  STAT_UPLOADS,
  STAT_ARCHIVES,
  STAT_SYMLINKS,
  STAT_RMSYMLINKS
};

ARGMATCH_VERIFY (stat_args, stat_types);

static void
_cfg_statmask (gsc_config_file_t *file, char *val, unsigned long *pmask)
{
  int i;
  int argc;
  char **argv;
  int invert = 0;
  unsigned long mask;
  
  if (argcv_get (val, NULL, NULL, &argc, &argv))
    {
      file->error_msg (file->file_name, file->line,
		       "cannot parse value: %s", strerror (errno));
      file->error_count++;
      return;
    }

  i = 0;
  if (strcmp (argv[0], "all") == 0)
    {
      mask = STAT_MASK_ALL;
      invert = 1;
      i++;
    }
  else if (strcmp (val, "none") == 0)
    {
      mask = STAT_MASK_NONE;
      i++;
    }
  else
    mask = STAT_MASK_NONE;

  for (; i < argc; i++)
    {
      ptrdiff_t x = ARGMATCH (argv[i], stat_args, stat_types);
      
      if (x == (ptrdiff_t)-1)
	{
	  file->error_msg (file->file_name, file->line,
			   "unknown statistics type: %s", argv[i]);
	  file->error_count++;
	  return;
	}
      if (invert)
	mask &= ~ STAT_MASK (stat_types[x]);
      else
	mask |= STAT_MASK (stat_types[x]);
    }
  *pmask = mask;
}

static void
cfg_statistics (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  _cfg_statmask (file, val, &print_stats);
}


static void
_cfg_email (gsc_config_file_t *file, char *kw, char *val, mu_address_t *paddr)
{
  mu_address_t addr;
  int rc = mu_address_create (&addr, val);
  if (rc)
    {
      file->error_msg (file->file_name, file->line,
		       "invalid mail address: %s", mu_strerror (rc));
      file->error_count++;
    }
  *paddr = addr;
}
    
static void
cfg_mailer (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  int rc = mu_mailer_create (&mailer, val);
  if (rc)
    {
      file->error_msg (file->file_name, file->line,
		       "creating mailer `%s' failed: %s",
		       val, mu_strerror (rc));
      file->error_count++;
    } 
}

static void
cfg_admin_address (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  _cfg_email (file, kw, val, &admin_address);
}

static void
cfg_from_address (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  _cfg_email (file, kw, val, &from_address);
}

static void
cfg_mail_admin_stat (gsc_config_file_t *file, char *kw, char *val,
		     void *unused)
{
  char *word;

  if (!val)
    {
      file->error_msg (file->file_name, file->line,
		       "missing message template identifier");
      file->error_count++;
      return;
    }

  word = get_word (&val);

  admin_stat_message_template = find_message_template (word);
  if (!admin_stat_message_template)
    {
      file->error_msg (file->file_name, file->line,
		       "no such message template: %s", word);
      file->error_count++;
      return;
    }
  _cfg_statmask (file, val, &mail_admin_mask);
}

static int
end_marker_p (const char *buf, const char *word, size_t wlen)
{
  if (strlen (buf) >= wlen && memcmp (buf, word, wlen) == 0)
    {
      while (buf[wlen] && isspace (buf[wlen]))
	wlen++;
      if (buf[wlen] == 0 || buf[wlen] == '#')
	return 1;
    }
  return 0;
}

static int
_cfg_recursive_raw_read (gsc_config_file_t *file, struct obstack *stk,
			 char *word, size_t wlen, int skip_initial_ws)
{
  while (getline (&file->buf, &file->size, file->fp) > 0)
    {
      char *p = file->buf;
      
      file->line++;
      
      if (skip_initial_ws)
	skip_hws (p);

      if (end_marker_p (p, word, wlen))
	{
	  obstack_1grow (stk, 0);
	  return 0;
	}
      else if (strlen (p) > 7 && memcmp (p, "include", 7) == 0
	       && (p[7] == ' ' || p[8] == '\t'))
	{
	  gsc_config_file_t new_file;
	  
	  p += 7;
	  skip_hws (p);
	  trim (p);
	  if (gsc_config_include_start (&new_file, file, p) == 0)
	    {
	      int rc = _cfg_recursive_raw_read (&new_file, stk,
						word, wlen, skip_initial_ws);
	      gsc_config_include_stop (&new_file);
	      if (rc == 0)
		return 0;
	    }
	}
      else
	{
	  if (*p == '>')
	    p++;
	  obstack_grow (stk, p, strlen (p));
	}
    }
  return 1;
}

static int
_cfg_raw_read (gsc_config_file_t *file, char *val, struct obstack *stk)
{
  char *word = val ? get_word (&val) : "";
  size_t wlen;
  size_t start_line = file->line;
  int skip_initial_ws = 0;
  
  if (word[0] == 0)
    word = "end";
  if (word[0] == '-')
    {
      skip_initial_ws = 1;
      word++;
    }
  wlen = strlen (word);

  if (_cfg_recursive_raw_read (file, stk, word, wlen, skip_initial_ws))
    {
      file->error_msg (file->file_name, start_line, "missing end marker");
      file->error_count++;
      return 1;
    }
  return 0;
}

static void
cfg_define_message (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  char *word;
  if (!val)
    {
      file->error_msg (file->file_name, file->line,
		       "missing message identifier");
      file->error_count++;
      return;
    }

  word = xstrdup (get_word (&val));
  if (_cfg_raw_read (file, val, &cfg_stk) == 0)
    register_message_template (word, obstack_finish (&cfg_stk));
}


int
string_to (char *what, char *str,
	   const char **args, int *vals,
	   int *pret,
	   gsc_config_file_t *file)
{
  ptrdiff_t x = ARGMATCH (str, args, vals);
      
  if (x == (ptrdiff_t)-1)
    {
      file->error_msg (file->file_name, file->line,
		       "unknown %s: %s", what, str);
      file->error_count++;
      return 1;
    }
  *pret = vals[x];
  return 0;
}

static const char *event_args[] = {
  "success",
  "bad-ownership",
  "bad-directive-signature",
  "bad-detached-signature",
  NULL
};

static int event_types[] = {
  ev_success,
  ev_bad_ownership,
  ev_bad_directive_signature,
  ev_bad_detached_signature
};

ARGMATCH_VERIFY (event_args, event_types);

const char *
notification_event_str (enum notification_event evt)
{
  return event_args[evt];
}

int
string_to_notification_event (gsc_config_file_t *file, char *val,
			      enum notification_event *pret)
{
  int rc, res;
  rc = string_to ("notification event", val,
		  event_args, event_types,
		  &res,
		  file);
  *pret = res;
  return rc;
}

static const char *target_args[] = {
  "admin",
  "user",
  "owner",
  NULL
};

static int target_types[] = {
  notify_admin,        /* System administrator */
  notify_owner,        /* Project admin */
  notify_user          /* User (uploader) */
};

ARGMATCH_VERIFY (target_args, target_types);

const char *
notification_target_str (enum notification_target tgt)
{
  return target_args[tgt];
}

int
string_to_notification_target (gsc_config_file_t *file, char *val,
			       enum notification_target *pret)
{
  int rc, res;
  rc = string_to ("notification target", val,
		  target_args, target_types,
		  &res,
		  file);
  *pret = res;
  return rc;
}

static void
cfg_notify_event (gsc_config_file_t *file, char *kw, char *val, void *unused)
{
  int argc;
  char **argv;
  enum notification_event evt;
  enum notification_target tgt;
  struct message_template *msg;
  int rc;
    
  if (argcv_get (val, NULL, NULL, &argc, &argv))
    {
      file->error_msg (file->file_name, file->line,
		       "cannot parse value: %s", strerror (errno));
      file->error_count++;
      return;
    }

  if (argc != 3)
    {
      file->error_msg (file->file_name, file->line,
		       "wrong number of arguments");
      file->error_count++;
      argcv_free (argc, argv);
      return;
    }

  rc = string_to_notification_event (file, argv[0], &evt)
       | string_to_notification_target (file, argv[1], &tgt);
  if ((msg = find_message_template (argv[2])) == NULL)
    {
      rc = 1;
      file->error_msg (file->file_name, file->line,
		       "undefined message: %s", argv[2]);
      file->error_count++;
    }

  if (rc == 0)
    register_notification (evt, tgt, msg);

  argcv_free (argc, argv);
}

static void
cfg_enable_sendfile (gsc_config_file_t *file, char *kw, char *val,
		     void *unused)
{
  if (get_bool (val, &enable_sendfile))
    {
      file->error_msg (file->file_name, file->line,
		       "boolean value expected, but found `%s'", val);
      file->error_count++;
    }
}

static void
cfg_archive_signatures (gsc_config_file_t *file, char *kw, char *val,
			void *unused)
{
  if (get_bool (val, &archive_signatures))
    {
      file->error_msg (file->file_name, file->line,
		       "boolean value expected, but found `%s'", val);
      file->error_count++;
    }
}



static struct gsc_config_keyword kw_handler[] = {
  { "include", gsc_config_include_handler },
  { "syslog-tag", cfg_syslog_tag },
  { "syslog-facility", cfg_syslog_facility },
  { "syslog-print-priority", cfg_syslog_print_priority },
  { "statistics", cfg_statistics },
  { "directory", cfg_directory },
  { "file-sweep-time", cfg_file_sweep_time },
  { "sql", cfg_sql },
  { "verify-user", cfg_verify_user },
  { "gpg-key", cfg_gpg_key },
  { "project-owner", cfg_project_owner },
  { "user-data", cfg_user_data },
  { "umask", cfg_umask },
  { "tar-program", cfg_tar_program },
  { "archive", cfg_archive },
  { "mailer", cfg_mailer },
  { "admin-address", cfg_admin_address },
  { "from-address", cfg_from_address },
  { "mail-admin-stat", cfg_mail_admin_stat },
  { "define-message", cfg_define_message },
  { "notify-event", cfg_notify_event },
  { "enable-sendfile", cfg_enable_sendfile },
  { "archive-signatures", cfg_archive_signatures },
  { NULL }
};

void
parse_config ()
{
  default_verify_method.type = method_builtin;
  default_gpg_key_method.type = method_builtin;
  default_project_owner_method.type = method_none;
  default_user_data_method.type = method_none;
  default_archive_descr.type = archive_none;
  default_archive_descr.backup_type = no_backups;

  obstack_init (&cfg_stk);
  switch (gsc_config_parse (conffile, NULL, kw_handler))
    {
    case gsc_config_success:
      if (mail_admin_mask)
	{
	  if (!admin_stat_message_template)
	    error (1, 0,
		   "mail-admin-stat given, but not admin-stat-message");
	  if (!admin_address)
	    error (1, 0, "mail-admin-stat given, but not admin-address");
	}
      break;

    case gsc_config_open:
      error (1, errno, "cannot open config file `%s'", conffile);

    case gsc_config_error:
      exit (1);
    }
}

